///////////////////////////////////////////////////////////////////////////////
// File: vicinity.cpp                                                        //
// Description: source file for particle vicinity (Cvicinity class)          //
// This file is part of the SISCone project.                                 //
// For more details, see http://www.lpthe.jussieu.fr/~soyez/siscone          //
//                                                                           //
// Copyright (c) 2006 Gavin Salam and Gregory Soyez                          //
//                                                                           //
// This program is free software; you can redistribute it and/or modify      //
// it under the terms of the GNU General Public License as published by      //
// the Free Software Foundation; either version 2 of the License, or         //
// (at your option) any later version.                                       //
//                                                                           //
// This program is distributed in the hope that it will be useful,           //
// but WITHOUT ANY WARRANTY; without even the implied warranty of            //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             //
// GNU General Public License for more details.                              //
//                                                                           //
// You should have received a copy of the GNU General Public License         //
// along with this program; if not, write to the Free Software               //
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA //
//                                                                           //
// $Revision:: 151                                                          $//
// $Date:: 2007-01-25 15:51:01 +0100 (Thu, 25 Jan 2007)                     $//
///////////////////////////////////////////////////////////////////////////////

#include "vicinity.h"
#include <math.h>
#include <algorithm>
#include <iostream>

namespace siscone{

using namespace std;

/*************************************************************
 * Cvicinity_elm implementation                              *
 * element in the vicinity of a parent.                      *
 * class used to manage one points in the vicinity           *
 * of a parent point.                                        *
 *************************************************************/

// ordering pointers to Cvicinity_elm
//------------------------------------
bool ve_less(Cvicinity_elm *ve1, Cvicinity_elm *ve2){
  return ve1->angle < ve2->angle;
}


/*************************************************************
 * Cvicinity implementation                                  *
 * list of element in the vicinity of a parent.              *
 * class used to manage the points which are in the vicinity *
 * of a parent point. The construction of the list can be    *
 * made from a list of points or from a quadtree.            *
 *************************************************************/

// default constructor
//---------------------
Cvicinity::Cvicinity(){
  n_part = 0;

  ve_list = NULL;
  quadtree = NULL;

  parent = NULL;
  VR2 = VR = 0.0;

  pi2 = 2.0*M_PI;
}

// constructor with initialisation
//---------------------------------
Cvicinity::Cvicinity(vector<Cmomentum> &_particle_list){
  quadtree = NULL;
  parent = NULL;
  VR2 = VR = 0.0;

  set_particle_list(_particle_list);

  pi2 = 2.0*M_PI;
}

// default destructor
//--------------------
Cvicinity::~Cvicinity(){
  if (ve_list!=NULL)
    delete[] ve_list;

  if (quadtree!=NULL)
    delete quadtree;
}

/*
 * set the particle_list
 *  - particle_list   list of particles (type Cmomentum)
 *  - n               number of particles in the list
 ************************************************************/ 
void Cvicinity::set_particle_list(vector<Cmomentum> &_particle_list){
  int i,j;
  double eta_max;
  
  // if the particle list is not empty, destroy it !
  if (quadtree!=NULL){
    delete quadtree;
  }
  if (ve_list!=NULL){
    delete[] ve_list;
  }
  vicinity.clear();
    
  // allocate memory array for particles
  // Note: - we compute max for |eta|
  //       - we allocate indices to particles
  n_part = 0;
  eta_max = 0.0;
  plist.clear();
  for (i=0;i<(int) _particle_list.size();i++){
    // if a particle is colinear with the beam (infinite rapidity)
    // we do not take it into account
    if (fabs(_particle_list[i].pz)!=_particle_list[i].E){
      plist.push_back(_particle_list[i]);
      plist[n_part].parent_index = i;
      plist[n_part].index = n_part;
      if (fabs(plist[n_part].eta)>eta_max) eta_max=fabs(plist[n_part].eta);
      // make sure the reference is randomly created
      plist[n_part].ref.randomize();

      n_part++;
    }
  }
  eta_max+=0.1;

  // allocate quadtree and vicinity_elm list
  // note: we set phi in [-pi:pi] as it is the natural range for atan2!
#ifdef USE_QUADTREE
  quadtree = new Cquadtree(0.0, 0.0, eta_max, M_PI);
#endif
  ve_list = new Cvicinity_elm[2*n_part];

  // append particle to the quadtree and to the vicinity_elm list
  j = 0;
  for (i=0;i<n_part;i++){
#ifdef USE_QUADTREE
    quadtree->add(&plist[i]);
#endif
    ve_list[j].v = ve_list[j+1].v = &plist[i];
    j+=2;
  }

}


/*
 * build the vicinity list from a quadtree.
 *  - _parent   reference particle
 *  - _VR       vicinity radius
 ************************************************************/
void Cvicinity::build_from_quadtree(Cmomentum *_parent, double _VR){

  // set parent and radius
  parent = _parent;
  VR  = _VR;
  VR2 = VR*VR;

  // clear vicinity
  vicinity.clear();

  // init parent variables
  pcx = parent->eta;
  pcy = parent->phi;

  // really browse the quadtree
  build_from_quadtree_internal(quadtree);

  // sort the vicinity
  sort(vicinity.begin(), vicinity.end(), ve_less);

  vicinity_size = vicinity.size();
}

/*
 * build the vicinity list from a list of points.
 *  - _parent   reference particle
 *  - _VR       vicinity radius
 ************************************************************/
void Cvicinity::build_from_list(Cmomentum *_parent, double _VR){
  int i;

  // set parent and radius
  parent = _parent;
  VR  = _VR;
  VR2 = VR*VR;

  // clear vicinity
  vicinity.clear();

  // init parent variables
  pcx = parent->eta;
  pcy = parent->phi;

  // really browse the particle list
  for (i=0;i<n_part;i++){
    append_to_vicinity(&plist[i]);
  }

  // sort the vicinity
  sort(vicinity.begin(), vicinity.end(), ve_less);

  vicinity_size = vicinity.size();
}


/*
 * internal tool to build the vicinity list from a quadtree.
 * \param qt    quadtree to look into
 *************************************************************/
void Cvicinity::build_from_quadtree_internal(Cquadtree *qt){
  // if no particle at all, return
  if (qt->v==NULL)
    return;

  // if 1! partic, test it and return
  if (!qt->has_child){
    append_to_vicinity(qt->v);
    return;
  }

  // check intersection
  // if intersection, forward to childs
  double rx, ry;

  // step 3: check if there is an intersection
  // compute (minimal) the distance (pay attention to the periodicity in phi).
  rx = fabs(pcx - qt->centre_x) - qt->half_size_x;
  if (rx<0) rx = 0.0;

  ry = fabs(pcy - qt->centre_y);
  if (ry>M_PI) ry = pi2-ry;
  ry -= qt->half_size_y;
  if (ry<0) ry = 0.0;

  // check the distance 
  if (rx*rx+ry*ry<=VR2){
    build_from_quadtree_internal(qt->childs[0][0]);
    build_from_quadtree_internal(qt->childs[0][1]);
    build_from_quadtree_internal(qt->childs[1][0]);
    build_from_quadtree_internal(qt->childs[1][1]);
  }
} 


/// strictly increasing function of the angle 
inline double sort_angle(double s, double c){
  if (s==0) return (c>0) ? 0.0 : 2.0;
  double t=c/s;
  return (s>0) ? 1-t/(1+fabs(t)) : 3-t/(1+fabs(t));
}


/*
 * append a particle to the 'vicinity' list after
 * having computed the angular-ordering quantities
 *  - v   vector to test
 **********************************************************/
void Cvicinity::append_to_vicinity(Cmomentum *v){
  double dx, dy, d2;

  // skip the particle itself)
  if (v==parent)
    return;

  int i=2*(v->index);

  // compute the distance of the i-th particle with the parent
  dx = v->eta - pcx;
  dy = v->phi - pcy;
    
  // pay attention to the periodicity in phi !
  if (dy>M_PI) 
    dy -= pi2;
  else if (dy<-M_PI) 
    dy += pi2;

  d2 = dx*dx+dy*dy;
    
  // really check if the distance is less than VR
  if (d2<VR2){
    double s,c,tmp;
    
    // compute the angles used for future ordering ...
    //  - build temporary bariables used for the computation
    //d   = sqrt(d2);
    tmp = sqrt(VR2/d2-1);

    // first angle (+)
    c = 0.5*(dx-dy*tmp);  // cosine of (parent,child) pair w.r.t. horizontal
    s = 0.5*(dy+dx*tmp);  // sine   of (parent,child) pair w.r.t. horizontal
    ve_list[i].angle = sort_angle(s,c);
    ve_list[i].eta = pcx+c;
    ve_list[i].phi = pcy+s;
    if (ve_list[i].phi<-M_PI) ve_list[i].phi+=pi2;
    if (ve_list[i].phi> M_PI) ve_list[i].phi-=pi2;
    ve_list[i].side = true;
    vicinity.push_back(&(ve_list[i]));

    // second angle (-)    
    c = 0.5*(dx+dy*tmp);  // cosine of (parent,child) pair w.r.t. horizontal
    s = 0.5*(dy-dx*tmp);  // sine   of (parent,child) pair w.r.t. horizontal
    ve_list[i+1].angle = sort_angle(s,c);
    ve_list[i+1].eta = pcx+c;
    ve_list[i+1].phi = pcy+s;
    if (ve_list[i+1].phi<-M_PI) ve_list[i+1].phi+=pi2;
    if (ve_list[i+1].phi> M_PI) ve_list[i+1].phi-=pi2;
    ve_list[i+1].side = false;
    vicinity.push_back(&(ve_list[i+1]));
  }
}

}
